// Copyright (C) 2019  Wilko Manger
//
// This Source Code Form is subject to the terms of the Mozilla Public
// License, v. 2.0. If a copy of the MPL was not distributed with this
// file, You can obtain one at http://mozilla.org/MPL/2.0/.

import 'package:matrix_sdk/src/event/room/room_event.dart';
import 'package:matrix_sdk/src/event/room/state/state_event.dart';
import 'package:matrix_sdk/src/identifier.dart';

import '../../event.dart';

class PowerLevelsChangeEvent extends StateEvent {
  @override
  final PowerLevelsChange content;

  PowerLevelsChangeEvent(this.content, RoomEventArgs args)
      : super(args, stateKey: '');

  factory PowerLevelsChangeEvent.fromJson(RoomEventArgs args, dynamic json) {
    if (!hasStateKey(json)) {
      return null;
    }

    final content = json['content'];

    final int ban = content['ban'] ?? 50;
    final int invite = content['invite'] ?? 50;
    final int kick = content['kick'] ?? 50;
    final int redact = content['redact'] ?? 50;

    final int eventsDefault = content['events_default'] ?? 0;
    final jsonEvents = content['events'];

    Map<Type, int> events;

    if (jsonEvents != null) {
      jsonEvents.removeWhere((type, powerLevel) => Event.typeOf(type) == null);

      events = jsonEvents?.map((eventType, powerLevel) {
        powerLevel = powerLevel is String ? int.parse(powerLevel) : powerLevel;
        return MapEntry<Type, int>(Event.typeOf(eventType), powerLevel);
      })?.cast<Type, int>();
    }

    final int stateDefault = content['state_default'] ?? 50;

    final int userDefault = content['users_default'] ?? 0;
    final Map<UserId, int> users = content['users'].map((userId, powerLevel) {
      powerLevel = powerLevel is String ? int.parse(powerLevel) : powerLevel;
      return MapEntry(UserId(userId), powerLevel);
    }).cast<UserId, int>();

    final roomNotifications = json.containsKey('notifications')
        ? json['notifications']['room'] ?? 50
        : 50;

    return PowerLevelsChangeEvent(
      PowerLevelsChange(
        banLevel: ban,
        inviteLevel: invite,
        kickLevel: kick,
        redactLevel: redact,
        stateEventsDefaultLevel: stateDefault,
        eventsDefaultLevel: eventsDefault,
        eventLevels: events,
        userDefaultLevel: userDefault,
        userLevels: users,
        roomNotificationLevel: roomNotifications,
      ),
      args,
    );
  }
}

class PowerLevelsChange extends StateEventContent {
  final int banLevel;
  final int inviteLevel;
  final int kickLevel;
  final int redactLevel;

  final int stateEventsDefaultLevel;
  final int eventsDefaultLevel;
  final Map<Type, int> eventLevels;

  final int userDefaultLevel;
  final Map<UserId, int> userLevels;

  final int roomNotificationLevel;

  PowerLevelsChange({
    this.banLevel = 50,
    this.inviteLevel = 50,
    this.kickLevel = 50,
    this.redactLevel = 50,
    this.stateEventsDefaultLevel = 50,
    this.eventsDefaultLevel = 0,
    this.eventLevels,
    this.userDefaultLevel = 0,
    this.userLevels,
    this.roomNotificationLevel = 50,
  });

  @override
  Map<String, dynamic> toJson() => super.toJson()
    ..addAll({
      'ban': banLevel,
      'invite': inviteLevel,
      'kick': kickLevel,
      'redact': redactLevel,
      'state_default': stateEventsDefaultLevel,
      'events_default': eventsDefaultLevel,
      'events': eventLevels
          ?.map((type, power) => MapEntry(Event.matrixTypeOf(type), power)),
      'users_default': userDefaultLevel,
      'users':
          userLevels.map((userId, power) => MapEntry(userId.toString(), power)),
      'notifications': {
        'room': roomNotificationLevel,
      },
    });
}
