// Copyright (C) 2019  Wilko Manger
//
// This Source Code Form is subject to the terms of the Mozilla Public
// License, v. 2.0. If a copy of the MPL was not distributed with this
// file, You can obtain one at http://mozilla.org/MPL/2.0/.

import 'package:matrix_sdk/matrix_sdk.dart';
import 'package:matrix_sdk/src/event/room/room_event.dart';
import 'package:matrix_sdk/src/event/room/state/state_event.dart';
import 'package:matrix_sdk/src/identifier.dart';
import 'package:meta/meta.dart';
import 'package:url/url.dart';

import '../../../user.dart';

class RoomCreationEvent extends StateEvent {
  @override
  final RoomCreation content;

  RoomCreationEvent(this.content, RoomEventArgs args)
      : super(args, stateKey: '');

  static Future<RoomCreationEvent> fromJson(
    RoomEventArgs args,
    dynamic json, {
    ProcessContext context,
  }) async {
    if (!hasStateKey(json)) {
      return null;
    }

    final content = json['content'];

    final creatorId = UserId(content['creator']);

    // If the event is from storage, there will be creator state info available
    User creator = User(id: creatorId);
    final creatorState = content['creator_state'];
    if (creatorState != null) {
      var avatarUrl = creatorState['avatar_url'];
      if (avatarUrl != null) {
        avatarUrl = Url.parse(avatarUrl);
      }

      var since = creatorState['since'];
      if (since != null) {
        since = DateTime.fromMillisecondsSinceEpoch(since);
      }

      creator = User(
          id: creatorId,
          state: UserState(
            displayName: creatorState['display_name'],
            avatarUrl: avatarUrl,
            since: since,
            roomId: args.roomId,
          ));
    }

    final federate = content['m.federate'] ?? true;
    final roomVersion = content['room_version'] ?? '1';

    var previousRoomId, previousRoomEventId;

    final predecessor = content['predecessor'];
    if (predecessor != null) {
      previousRoomId = predecessor['room_id'];
      if (previousRoomId != null) {
        previousRoomId = RoomId(previousRoomId);
      }

      previousRoomEventId = predecessor['event_id'];
      if (previousRoomEventId != null) {
        previousRoomEventId = EventId(previousRoomEventId);
      }
    }

    return RoomCreationEvent(
      RoomCreation(
        creator: creator ?? await context?.members?.getStateOf(creatorId),
        federate: federate,
        roomVersion: roomVersion,
        previousRoomId: previousRoomId,
        previousRoomLastEventId: previousRoomEventId,
      ),
      args,
    );
  }
}

class RoomCreation extends StateEventContent {
  /// User id of the creator of the room.
  ///
  /// Use [room.creator] to get the state of the
  /// creator at time of creation.
  final User creator;

  /// Whether the created room is federated.
  final bool federate;

  final String roomVersion;

  final RoomId previousRoomId;
  final EventId previousRoomLastEventId;

  RoomCreation({
    @required this.creator,
    @required this.federate,
    @required this.roomVersion,
    @required this.previousRoomId,
    @required this.previousRoomLastEventId,
  });

  @override
  Map<String, dynamic> toJson() => super.toJson()
    ..addAll({
      'm.federate': federate,
      'room_version': roomVersion,
      'creator': creator.id,
      'predecessor': {
        'room_id': previousRoomId?.toString(),
        'event_id': previousRoomLastEventId?.toString(),
      },
      // Not a official field in the spec
      'creator_state': {
        'display_name': creator.name,
        'avatar_url': creator.avatarUrl?.toString(),
        'since': creator.state?.since?.millisecondsSinceEpoch,
      }
    });
}
