// Copyright (C) 2019  Wilko Manger
// Copyright (C) 2019  Mathieu Velten
//
// This Source Code Form is subject to the terms of the Mozilla Public
// License, v. 2.0. If a copy of the MPL was not distributed with this
// file, You can obtain one at http://mozilla.org/MPL/2.0/.

import 'dart:async';

import 'package:matrix_sdk/matrix_sdk.dart';
import 'package:matrix_sdk/src/event/event.dart';
import 'package:matrix_sdk/src/event/room/room_event.dart';
import 'package:matrix_sdk/src/room/room.dart';
import 'package:matrix_sdk/src/storables.dart';
import 'package:matrix_sdk/src/identifier.dart';
import 'package:meta/meta.dart';

/// Sort order: Newest first based on [RoomEvent.time].
class Timeline extends Storables<EventId, RoomEvent> {
  final Room _room;

  Timeline(this._room, String scope) : super(scope);

  @override
  @protected
  Future<RoomEvent> getSingleFromStore(EventId id) =>
      store?.getRoomEvent(_room, id);

  @override
  @protected
  Future<Iterable<RoomEvent>> getFromStore({RoomEvent from, int count}) =>
      store?.getRoomEvents(_room, from: from, count: count);

  @override
  @protected
  Future<int> getSavedInStoreCount() => store?.getRoomEventCount(_room.id);

  @override
  @protected
  Future<void> setInStore(Iterable<RoomEvent> events) =>
      store?.addRoomEvents(_room, events);

  @protected
  @override
  Future<void> removeInStore(EventId id) async {
    await store?.removeEvent(id);
  }

  @protected
  @override
  Future<Iterable<RoomEvent>> getFromRemote({int count}) async =>
      _room.retrieveRemoteEvents(count);

  @override
  Iterable<RoomEvent> sort(Iterable<Event> items) {
    final list = items.toList(growable: false);
    list.sort((a, b) {
      if (a is RoomEvent && b is RoomEvent) {
        return a.time.compareTo(b.time);
      } else {
        return a.id.toString().compareTo(b.id.toString());
      }
    });

    return list.reversed;
  }

  @override
  bool isLastPossibleItem(RoomEvent item) => item is RoomCreationEvent;

  @override
  Future<void> addAll(Iterable<RoomEvent> values) async {
    // Remove echo events
    for (final event in values) {
      final transactionId = event?.unsigned?.transactionId;
      if (transactionId != null) {
        // delete local echo
        await remove(EventId(transactionId));
      }
    }

    await super.addAll(values);
  }
}
