// Copyright (C) 2019  Wilko Manger
//
// This Source Code Form is subject to the terms of the Mozilla Public
// License, v. 2.0. If a copy of the MPL was not distributed with this
// file, You can obtain one at http://mozilla.org/MPL/2.0/.

import 'dart:convert';

import 'package:chopper/chopper.dart';

/// Any possible error response that could be returned from
/// the API.
abstract class MatrixException implements Exception {
  /// Message that is sometimes included in the response.
  /// It's not recommended that you use this message to display to end
  /// users.
  String get message;

  /// The original json body.
  Map<String, dynamic> get body;

  @override
  String toString() => '${runtimeType}: ${json.encode(body)}';

  /// Throw an exception based on the given response.
  static void throwFromResponse(Response response) {
    var body;
    try {
      body = json.decode(response.body);
    } on FormatException {
      return;
    }

    final code = body['errcode'];
    final message = body['error'];

    switch (code) {
      case 'M_FORBIDDEN':
        throw ForbiddenException(message: message, body: body);
      case 'M_USER_IN_USE':
        throw UsernameInUseException(message: message, body: body);
      case 'M_INVALID_USERNAME':
        throw InvalidUsernameException(message: message, body: body);
      case 'M_EXCLUSIVE':
        throw ExclusiveException(message: message, body: body);
      case 'M_UNKNOWN':
        throw UnknownException(message: message, body: body);
    }
  }
}

/// Unknown error occured. Check the [message] for details.
class UnknownException extends MatrixException {
  UnknownException({this.message, this.body});

  @override
  final String message;

  @override
  final Map<String, dynamic> body;
}

/// Denied access. For example, wrong password, or not allowed
/// to join a room.
class ForbiddenException extends MatrixException {
  ForbiddenException({this.message, this.body});

  @override
  final String message;

  @override
  final Map<String, dynamic> body;
}

/// Username is already in use.
class UsernameInUseException extends MatrixException {
  UsernameInUseException({this.message, this.body});

  @override
  final String message;

  @override
  final Map<String, dynamic> body;
}

/// Invalid username.
class InvalidUsernameException extends MatrixException {
  InvalidUsernameException({this.message, this.body});

  @override
  final String message;

  @override
  final Map<String, dynamic> body;
}

/// The requested resource is reserved by an application service.
class ExclusiveException extends MatrixException {
  ExclusiveException({this.message, this.body});

  @override
  final String message;

  @override
  final Map<String, dynamic> body;
}
