// Copyright (C) 2019  Wilko Manger
// Copyright (C) 2019  Mathieu Velten
//
// This Source Code Form is subject to the terms of the Mozilla Public
// License, v. 2.0. If a copy of the MPL was not distributed with this
// file, You can obtain one at http://mozilla.org/MPL/2.0/.

import 'dart:convert';

import 'package:kiwi/kiwi.dart';
import 'package:matrix_sdk/src/api/client.dart';
import 'package:matrix_sdk/src/event/event.dart';
import 'package:matrix_sdk/src/event/room/state/aliases_change_event.dart';
import 'package:matrix_sdk/src/event/room/state/member_change_event.dart';
import 'package:matrix_sdk/src/event/room/state/power_levels_change_event.dart';
import 'package:matrix_sdk/src/event/room/state/room_creation_event.dart';
import 'package:matrix_sdk/src/event/room/state/room_upgrade_event.dart';
import 'package:matrix_sdk/src/event/room/state/state_event.dart';
import 'package:matrix_sdk/src/event/room/state/topic_change_event.dart';
import 'package:matrix_sdk/src/event/timeline.dart';
import 'package:matrix_sdk/src/event/room/redaction_event.dart';
import 'package:matrix_sdk/src/event/room/room_event.dart';
import 'package:matrix_sdk/src/event/room/state/room_avatar_change_event.dart';
import 'package:matrix_sdk/src/event/room/state/room_name_change_event.dart';
import 'package:matrix_sdk/src/identifier.dart';
import 'package:matrix_sdk/src/members.dart';
import 'package:matrix_sdk/src/user.dart';
import 'package:chopper/chopper.dart';
import 'package:matrix_sdk/src/util/random.dart';
import 'package:meta/meta.dart';
import 'package:url/url.dart';

import '../exception.dart';
import '../store.dart';

class Room extends Identifiable<RoomId> {
  @override
  final RoomId id;

  Timeline _timeline;
  Timeline get timeline => _timeline;

  Members _members;
  Members get members => _members;

  //// @internal
  String previousBatch;

  final String _scope;
  ClientService _client;

  RoomStateEvents _stateEvents;

  /// The latest [StateEvent]s of this room.
  RoomStateEvents get stateEvents => _stateEvents;

  String get name => stateEvents.nameChange?.content?.name;

  Url get avatarUrl => stateEvents.avatarChange?.content?.url;

  bool get isDirect => directUser != null;

  /// [UserId] of the user that created this room.
  User get creator => stateEvents.creation?.content?.creator;

  /// The [RoomId] of the room that replaced this room.
  RoomId get replacementId => stateEvents.upgrade?.content?.replacementRoomId;

  /// Whether this room is upgraded to another room.
  ///
  /// See [replacementId] to see which room this was upgraded.
  bool get isUpgraded => replacementId != null;

  /// The [RoomId] of the room that was replaced by this room.
  RoomId get replacesId => stateEvents.creation.content.previousRoomId;

  /// Whether this room is a replacement for another room.
  ///
  /// See [replacesId] to see which room was replaced by this room.
  bool get isReplacement => replacesId != null;

  String get topic => stateEvents.topicChange?.content?.topic;

  PowerLevels _powerLevels;

  /// Power levels for this room to do a certain action or send
  /// certain events.
  ///
  /// Also defaults and info of which user has which power level.
  PowerLevels get powerLevels => _powerLevels;

  List<RoomAlias> get aliases => stateEvents.aliasChanges.values
      .map((e) => e.content.aliases)
      .where((a) => a != null)
      .expand((a) => a)
      .toList(growable: false);

  List<User> _typingUsers = List();

  List<User> get typingUsers => _typingUsers;

  int _highlightedUnreadNotificationCount = 0;

  int get highlightedUnreadNotificationCount =>
      _highlightedUnreadNotificationCount;

  int _totalUnreadNotificationCount = 0;

  int get totalUnreadNotificationCount => _totalUnreadNotificationCount;

  final LocalUser _me;

  final Store _store;

  Room({
    @required this.id,
    @required String scope,
    RoomStateEvents stateEvents,
    int memberCount,
    User directUser,
    String previousBatch,
    int highlightedUnreadNotificationCount,
    int totalUnreadNotificationCount,
  })  : _scope = scope,
        _stateEvents = stateEvents ?? RoomStateEvents(),
        _directUser = directUser,
        previousBatch = previousBatch,
        _highlightedUnreadNotificationCount =
            highlightedUnreadNotificationCount ?? 0,
        _totalUnreadNotificationCount = totalUnreadNotificationCount ?? 0,
        _me = Container().resolve(scope),
        _store = Container().resolve(scope) {
    _members = Members(this, memberCount, scope);
    _timeline = Timeline(this, scope);
    _client = Container().resolve(_scope);
    _powerLevels = PowerLevels(_stateEvents.powerLevelsChange);
  }

  User _directUser;

  /// The other user of the direct chat. Is null if [isDirect] is false.
  User get directUser => _directUser;

  /// @internal
  Future<Iterable<User>> retrieveRemoteMembers() async {
    final response = await _client.members(
      roomId: id.toString(),
      at: previousBatch,
      membership: 'join',
    );

    final body = json.decode(response.body);

    final context = ProcessContext(members: members);

    await _processEvents(body['chunk'].reversed, context);

    return context.currentUsers.values;
  }

  /// @internal
  Future<Iterable<RoomEvent>> retrieveRemoteEvents(int limit) async {
    final response = await _client.roomMessages(
      roomId: id.toString(),
      limit: limit,
      from: previousBatch,
      filter: json.encode({
        'lazy_load_members': true,
      }),
    );

    final body = json.decode(response.body);

    previousBatch = body['end'];
    await _store.setRooms([this]);

    final context = ProcessContext(members: members);

    if (body.containsKey('state')) {
      await _processEvents(body['state'], context);
    }

    // TODO: When most event types are supported, don't check if they're
    // null anymore (null is now returned if the event type is unsupported)
    final processed = (await _processEvents(body['chunk'].reversed, context))
        .where((event) => event != null);

    return processed;
  }

  /// @internal
  Future<void> processSync(Map<String, dynamic> response) async {
    // Only set previous batch from sync when none is set
    if (previousBatch == null) {
      previousBatch = response['timeline']['prev_batch'];
    }

    final summary = response['summary'];
    if (summary != null) {
      members.count = summary['m.joined_member_count'] ?? members.count;
    }

    final notifications = response['unread_notifications'];
    if (notifications != null) {
      _highlightedUnreadNotificationCount =
          notifications['highlight_count'] ?? 0;
      _totalUnreadNotificationCount = notifications['notification_count'] ?? 0;
    }

    final context = ProcessContext(members: members);

    await _processEvents(response['state']['events'], context);

    final processed = await _processEvents(
      response['timeline']['events'],
      context,
    );

    await timeline.addAll(processed);
  }

  Future<List<RoomEvent>> _processEvents(
      Iterable<dynamic> rawEvents, ProcessContext context) async {
    bool isNewer(RoomEvent a, RoomEvent b) {
      return (a?.time?.millisecondsSinceEpoch ?? 0) >=
          (b?.time?.millisecondsSinceEpoch ?? 0);
    }

    final processedEvents = List<RoomEvent>();
    for (final rawEvent in rawEvents) {
      final event = await RoomEvent.fromJson(
        RoomEventArgs(roomId: id),
        rawEvent,
        context: context,
      );

      processedEvents.add(event);

      if (event is RedactionEvent) {
        final redactedId = event.redacts;

        final original = await timeline[redactedId];
        if (original != null && original is! RedactedEvent) {
          await timeline.remove(redactedId);

          await timeline.addAll([
            RedactedEvent.from(
              redaction: event,
              original: original,
            )
          ]);
        }
      } else if (event is RoomNameChangeEvent &&
          isNewer(event, stateEvents.nameChange)) {
        var name = event.content.name;
        // Set to null if name is empty
        if (name != null && name.isEmpty) {
          name = null;
        }

        _stateEvents = stateEvents.copyWith(nameChangeEvent: event);
      } else if (event is RoomAvatarChangeEvent &&
          isNewer(event, stateEvents.avatarChange)) {
        _stateEvents = stateEvents.copyWith(avatarChangeEvent: event);
      } else if (event is AliasesChangeEvent &&
          (stateEvents.aliasChanges[event.stateKey] == null ||
              isNewer(event, stateEvents.aliasChanges[event.stateKey]))) {
        final aliases = Map.of(stateEvents.aliasChanges);

        aliases[event.stateKey] = event;

        _stateEvents = _stateEvents.copyWith(aliasesChangeEvents: aliases);
      } else if (event is PowerLevelsChangeEvent &&
          isNewer(event, stateEvents.powerLevelsChange)) {
        _stateEvents = stateEvents.copyWith(powerLevelsChangeEvent: event);
        _powerLevels = PowerLevels(stateEvents.powerLevelsChange);
      } else if (event is RoomCreationEvent) {
        _stateEvents = stateEvents.copyWith(creationEvent: event);
      } else if (event is RoomUpgradeEvent &&
          isNewer(event, stateEvents.upgrade)) {
        _stateEvents = stateEvents.copyWith(upgradeEvent: event);
      } else if (event is TopicChangeEvent &&
          isNewer(event, stateEvents.topicChange)) {
        _stateEvents = stateEvents.copyWith(topicChangeEvent: event);
      } else if (event is InviteEvent && event.content.isDirect) {
        // Check whether there are only two people in the
        // room
        if (members.count <= 2) {
          if (event.content.subject != _me) {
            _directUser = event.content.subject;
          } else if (event.sender != _me) {
            _directUser = event.sender;
          }
        }
      } else if (event is JoinEvent && this.isDirect) {
        // If someone joins who is not me or the direct user,
        // this room is no longer direct
        final subject = event.content.subject;
        if (subject != _me && subject != directUser) {
          _directUser = null;
        }
      }
    }

    await members.addAll(context.users);

    return processedEvents;
  }

  Future<void> processUserAccountData(Map<String, dynamic> raw) async {
    final events = raw['events'] as List<dynamic>;

    if (_directUser == null) {
      final event = events.firstWhere(
        (event) => event['type'] == 'm.direct',
        orElse: () => null,
      );

      if (event != null) {
        final content = event['content'] as Map<String, dynamic>;

        await content.forEach((userId, roomIds) async {
          if (UserId.isValidFullyQualified(userId) &&
              roomIds.contains(id.toString())) {
            // Check whether there are only two people in the
            // room
            if (members.count <= 2) {
              _directUser = await members[UserId(userId)];
            }
          }
        });
      }
    }
  }

  Future<void> processEphemeral(Map<String, dynamic> raw) async {
    if (raw['ephemeral'] == null || raw['ephemeral']['events'] == null) {
      return;
    }

    final events = raw['ephemeral']['events'] as List<dynamic>;

    final rawTypingEvent = events.firstWhere(
      (rawEvent) => rawEvent['type'] == 'm.typing',
      orElse: () => null,
    );

    if (rawTypingEvent == null) return;

    final typingUsers = (await Future.wait(rawTypingEvent['content']['user_ids']
            .map((id) => Future.sync(() => members[UserId(id)]))
            .cast<Future<User>>()))
        .cast<User>();

    _typingUsers = typingUsers;
  }

  /// Whether someone (not you) is typing.
  bool get isSomeoneElseTyping =>
      typingUsers.isNotEmpty && !typingUsers.any((user) => user == _me);

  /// Whether someone is typing.
  bool get isSomeoneTyping => typingUsers.isNotEmpty;
}

class ProcessContext {
  final Map<EventId, RoomEvent> timeline = Map();

  /// User which is currently the correct one.
  final Map<UserId, User> currentUsers = Map();

  /// All user states.
  final List<User> users = List();

  /// Members instance to get users from if
  /// not in the context.
  final Members members;

  ProcessContext({@required this.members});

  void setUser(User user) {
    users.add(user);
    currentUsers[user.id] = user;
  }
}

class JoinedRoom extends Room {
  JoinedRoom({
    @required RoomId id,
    @required String scope,
    RoomStateEvents stateEvents,
    int memberCount,
    User directUser,
    String previousBatch,
    int highlightedUnreadNotificationCount,
    int totalUnreadNotificationCount,
  }) : super(
          id: id,
          scope: scope,
          stateEvents: stateEvents,
          memberCount: memberCount,
          directUser: directUser,
          previousBatch: previousBatch,
          highlightedUnreadNotificationCount:
              highlightedUnreadNotificationCount,
          totalUnreadNotificationCount: totalUnreadNotificationCount,
        );

  /// Send an [Event] to the [Room]. Can be any [RoomEvent]
  /// including [StateEvent]s.
  ///
  /// However, it's recommended to use the
  /// methods in [Room] to alter the state.
  ///
  /// Returns the up to date [SentState] of the sent event. (Not
  /// including [SentState.received].
  Stream<SentState> send(EventContent content,
      {String transactionId, String stateKey = ''}) async* {
    if (transactionId == null) {
      transactionId = randomString();
    }

    final eventArgs = RoomEventArgs(
      id: EventId(transactionId),
      roomId: id,
      time: DateTime.now(),
      sender: User(
        id: Container().resolve<LocalUser>(_scope).id,
      ),
      unsigned: UnsignedData(transactionId: transactionId),
      sentState: SentState.unsent,
    );

    final event = Event.fromContent(content, eventArgs);

    if (event == null || event is RoomCreationEvent) {
      throw ArgumentError('this event type cannot be send');
    }

    await timeline.addAll([event]);
    yield SentState.unsent;

    var sent = false;
    while (!sent) {
      Response response;

      try {
        final roomIdString = id.toString();
        final eventType = event.matrixType;
        final encodedContent = json.encode(content.toJson());

        if (event is StateEvent) {
          response = await _client.sendState(
            roomId: roomIdString,
            eventType: eventType,
            stateKey: stateKey,
            content: encodedContent,
          );
        } else {
          response = await _client.send(
            roomId: roomIdString,
            eventType: eventType,
            txnId: transactionId,
            content: encodedContent,
          );
        }

        final body = json.decode(response.body);
        final eventId = EventId(body['event_id']);

        final sentEvent = Event.fromContent(
          content,
          eventArgs.copyWith(
            id: eventId,
            sentState: SentState.sent,
          ),
        );

        sent = true;
        await timeline.addAll([sentEvent]);
        yield SentState.sent;
      } on Exception {
        yield SentState.failed;
        continue;
      }
    }
  }

  Future<void> setName(String name) =>
      send(RoomNameChange(name: name)).forEach((_) => null);

  Future<void> setAvatarUrl(Url avatarUrl) =>
      send(RoomAvatarChange(url: avatarUrl)).forEach((_) => null);

  Future<void> setTopic(String topic) =>
      send(TopicChange(topic: topic)).forEach((_) => null);

  Future<void> upgrade({
    @required RoomId replacementRoomId,
    String message,
  }) =>
      send(RoomUpgrade(
        replacementRoomId: replacementRoomId,
        body: message,
      )).forEach((_) => null);

  /// Notify whether the user is typing.
  ///
  /// If [typing] is true, a [timeout] can be specified
  /// to notify how long the server should assume that the user is
  /// typing.
  Future<void> setIsTyping(
    bool typing, {
    Duration timeout = const Duration(seconds: 30),
  }) async {
    final Map<String, dynamic> body = {
      'typing': typing,
    };

    if (typing) {
      body['timeout'] = timeout.inMilliseconds;
    }

    await _client.typing(
      roomId: id.toString(),
      userId: _me.id.toString(),
      body: json.encode(body),
    );
  }

  /// Will mark all messages as read until [until]. Will also
  /// send a read receipt, if [receipt] is true (default).
  Future<void> markRead({@required EventId until, bool receipt = true}) async {
    final body = {
      'm.fully_read': until.toString(),
    };

    if (receipt) {
      body['m.read'] = until.toString();
    }

    final response = await _client.readMarkers(
      roomId: id.toString(),
      body: json.encode(body),
    );

    MatrixException.throwFromResponse(response);
  }
}

class LeftRoom extends Room {
  LeftRoom({
    @required RoomId id,
    @required String scope,
    RoomStateEvents stateEvents,
    int memberCount,
    User directUser,
    String previousBatch,
    int highlightedUnreadNotificationCount,
    int totalUnreadNotificationCount,
  }) : super(
          id: id,
          scope: scope,
          stateEvents: stateEvents,
          memberCount: memberCount,
          directUser: directUser,
          previousBatch: previousBatch,
          highlightedUnreadNotificationCount:
              highlightedUnreadNotificationCount,
          totalUnreadNotificationCount: totalUnreadNotificationCount,
        );
}

/// Class that contains the most recent state events of each type.
class RoomStateEvents {
  final RoomNameChangeEvent nameChange;
  final RoomAvatarChangeEvent avatarChange;
  final TopicChangeEvent topicChange;
  final PowerLevelsChangeEvent powerLevelsChange;
  final RoomCreationEvent creation;
  final RoomUpgradeEvent upgrade;

  /// Alias changes mapped to their `stateKey`.
  final Map<String, AliasesChangeEvent> aliasChanges;

  RoomStateEvents({
    this.nameChange,
    this.avatarChange,
    this.topicChange,
    this.powerLevelsChange,
    this.aliasChanges = const {},
    this.creation,
    this.upgrade,
  });

  RoomStateEvents copyWith({
    RoomNameChangeEvent nameChangeEvent,
    RoomAvatarChangeEvent avatarChangeEvent,
    TopicChangeEvent topicChangeEvent,
    PowerLevelsChangeEvent powerLevelsChangeEvent,
    Map<String, AliasesChangeEvent> aliasesChangeEvents,
    RoomCreationEvent creationEvent,
    RoomUpgradeEvent upgradeEvent,
  }) {
    return RoomStateEvents(
      nameChange: nameChangeEvent ?? this.nameChange,
      avatarChange: avatarChangeEvent ?? this.avatarChange,
      topicChange: topicChangeEvent ?? this.topicChange,
      powerLevelsChange: powerLevelsChangeEvent ?? this.powerLevelsChange,
      aliasChanges: aliasesChangeEvents ?? this.aliasChanges,
      creation: creationEvent ?? this.creation,
      upgrade: upgradeEvent ?? this.upgrade,
    );
  }
}

class PowerLevels {
  int get ban => _content?.banLevel;
  int get invite => _content?.inviteLevel;
  int get kick => _content?.kickLevel;
  int get redact => _content?.redactLevel;

  int get roomNotification => _content?.roomNotificationLevel;

  Map<Type, int> get events => _content?.eventLevels;

  Map<UserId, int> get users => _content?.userLevels;

  final PowerLevelDefaults defaults;

  final PowerLevelsChange _content;

  PowerLevels(PowerLevelsChangeEvent event)
      : _content = event?.content,
        defaults = PowerLevelDefaults(event);
}

class PowerLevelDefaults {
  int get stateEvents => _content?.stateEventsDefaultLevel;
  int get events => _content?.eventsDefaultLevel;

  int get users => _content?.userDefaultLevel;

  final PowerLevelsChange _content;

  PowerLevelDefaults(PowerLevelsChangeEvent event) : _content = event?.content;
}
